HEADER_PEPTIDE = 'Peptide sequence'
HEADER_NO_PSM = 'PSM count'
HEADER_AREA = 'MS1 area (highest of all PSMs)'
HEADER_QVAL = 'q-value'
HEADER_PEP = 'PEP'
HEADER_QVAL_MODELED = 'q-value (linear modeled)'
HEADER_FALSE_LOC_RATE = 'PTM FLR'
HEADER_LINKED_PSMS = 'Files/scans for peptide'
HEADER_BAREPEP = 'Bare peptide'
HEADER_PROTEINS = 'Protein(s)'
HEADER_MASTERPROTEINS = 'Master protein(s)'
HEADER_STARTSTOP = 'Protein start/stop site(s)'
HEADER_NO_CONTENTPROTEINS = 'Total number proteins in group'
HEADER_COVERAGES = 'Coverage(s)'
HEADER_GENES = 'Gene ID(s)'
HEADER_ASSOCIATED = 'Gene name(s)'
HEADER_DESCRIPTIONS = 'Description(s)'
HEADER_SVM_SCORE = 'percolator svm-score'
